from enum import Enum


class ModelType(Enum):
    """A selection of machine learning modeling frameworks supported by Openlayer.

    .. note::
        Our `sample notebooks <https://github.com/openlayer-ai/openlayer-python/tree/main/examples>`_
        show you how to use each one of these model types with Openlayer.
    """

    #: For custom built models.
    custom = "custom"
    #: For models built with `fastText <https://fasttext.cc/>`_.
    fasttext = "fasttext"
    #: For models built with `Keras <https://keras.io/>`_.
    keras = "keras"
    #: For models built with `PyTorch <https://pytorch.org/>`_.
    pytorch = "pytorch"
    #: For models built with `rasa <https://rasa.com/>`_.
    rasa = "rasa"
    #: For models built with `scikit-learn <https://scikit-learn.org/>`_.
    sklearn = "sklearn"
    #: For models built with `TensorFlow <https://www.tensorflow.org/>`_.
    tensorflow = "tensorflow"
    #: For models built with `Hugging Face transformers <https://huggingface.co/docs/transformers/index>`_.
    transformers = "transformers"
    #: For models built with `XGBoost <https://xgboost.readthedocs.io>`_.
    xgboost = "xgboost"


class Model:
    """An object containing information about a model on the Openlayer platform."""

    def __init__(self, json):
        self._json = json
        self.id = json["id"]

    def __getattr__(self, name):
        if name in self._json:
            return self._json[name]
        raise AttributeError(f"'{type(self).__name__}' object has no attribute {name}")

    def __hash__(self):
        return hash(self.id)

    def __str__(self):
        return f"Model(id={self.id})"

    def __repr__(self):
        return f"Model({self._json})"

    def to_dict(self):
        """Returns object properties as a dict.

        Returns
        -------
        Dict with object properties.
        """
        return self._json



"""Schemas for the objects that shall be created on the Openlayer platform.
"""
import marshmallow as ma

from .datasets import DatasetType
from .models import ModelType
from .tasks import TaskType

# ---------------------------- Regular expressions --------------------------- #
COLUMN_NAME_REGEX = validate = ma.validate.Regexp(
    r"^[a-zA-Z0-9_-]+$",
    error="strings that are not alphanumeric with underscores or hyphens."
    + " Spaces and special characters are not allowed.",
)
LANGUAGE_CODE_REGEX = ma.validate.Regexp(
    r"^[a-z]{2}(-[A-Z]{2})?$",
    error="`language` of the dataset is not in the ISO 639-1 (alpha-2 code) format.",
)

# ---------------------------------- Schemas --------------------------------- #
class BaselineModelSchema(ma.Schema):
    """Schema for baseline models."""

    ensembleSize = ma.fields.Int(load_default=10)
    metadata = ma.fields.Dict(allow_none=True, load_default={})
    perRunLimit = ma.fields.Int(load_default=None, allow_none=True)
    randomSeed = ma.fields.Int(load_default=42)
    timeout = ma.fields.Int(load_default=60)


class CommitSchema(ma.Schema):
    """Schema for commits."""

    commitMessage = ma.fields.Str(
        required=True,
        validate=ma.validate.Length(
            min=1,
            max=140,
        ),
    )


class DatasetSchema(ma.Schema):
    """Schema for datasets."""

    categoricalFeatureNames = ma.fields.List(
        ma.fields.Str(validate=COLUMN_NAME_REGEX),
        allow_none=True,
        load_default=[],
    )
    classNames = ma.fields.List(ma.fields.Str(), required=True)
    columnNames = ma.fields.List(
        ma.fields.Str(validate=COLUMN_NAME_REGEX),
        required=True,
    )
    label = ma.fields.Str(
        validate=ma.validate.OneOf(
            [dataset_type.value for dataset_type in DatasetType],
            error="`label` must be one of the supported frameworks."
            + "Check out our API reference for a full list"
            + " https://reference.openlayer.com/reference/api/openlayer.DatasetType.html.\n ",
        ),
        required=True,
    )
    featureNames = ma.fields.List(
        ma.fields.Str(validate=COLUMN_NAME_REGEX),
        load_default=[],
    )
    labelColumnName = ma.fields.Str(
        validate=COLUMN_NAME_REGEX,
        required=True,
    )
    language = ma.fields.Str(
        load_default="en",
        validate=LANGUAGE_CODE_REGEX,
    )
    metadata = ma.fields.Dict(allow_none=True, load_default={})
    predictionsColumnName = ma.fields.Str(
        validate=COLUMN_NAME_REGEX,
        allow_none=True,
        load_default=None,
    )
    sep = ma.fields.Str(load_default=",")
    textColumnName = ma.fields.Str(
        validate=COLUMN_NAME_REGEX,
        allow_none=True,
    )

    @ma.validates_schema
    def validates_label_column_not_in_feature_names(self, data, **kwargs):
        """Validates whether the label column name is not on the feature names list"""
        if data["labelColumnName"] in data["featureNames"]:
            raise ma.ValidationError(
                f"`labelColumnName` `{data['labelColumnName']}` must not be in `featureNames`."
            )


class ModelSchema(ma.Schema):
    """Schema for models with artifacts (i.e., model_package)."""

    categoricalFeatureNames = ma.fields.List(
        ma.fields.Str(validate=COLUMN_NAME_REGEX),
        load_default=[],
    )
    classNames = ma.fields.List(
        ma.fields.Str(),
    )
    name = ma.fields.Str(
        required=True,
        validate=ma.validate.Length(
            min=1,
            max=64,
        ),
    )
    featureNames = ma.fields.List(
        ma.fields.Str(validate=COLUMN_NAME_REGEX),
        allow_none=True,
        load_default=[],
    )
    metadata = ma.fields.Dict(
        allow_none=True,
        load_default={},
    )
    architectureType = ma.fields.Str(
        validate=ma.validate.OneOf(
            [model_framework.value for model_framework in ModelType],
            error="`architectureType` must be one of the supported frameworks."
            + " Check out our API reference for a full list"
            + " https://reference.openlayer.com/reference/api/openlayer.ModelType.html.\n ",
        ),
        required=True,
    )


class ProjectSchema(ma.Schema):
    """Schema for projects."""

    description = ma.fields.Str(
        validate=ma.validate.Length(
            min=1,
            max=140,
        ),
    )
    name = ma.fields.Str(
        required=True,
        validate=ma.validate.Length(
            min=1,
            max=64,
        ),
    )
    task_type = ma.fields.Str(
        alidate=ma.validate.OneOf(
            [task_type.value for task_type in TaskType],
            error="`task_type` must be one of the supported tasks."
            + " Check out our API reference for a full list"
            + " https://reference.openlayer.com/reference/api/openlayer.TaskType.html.\n ",
        ),
    )


class ShellModelSchema(ma.Schema):
    """Schema for models without artifacts (i.e., shell model)."""

    name = ma.fields.Str(
        required=True,
        validate=ma.validate.Length(
            min=1,
            max=64,
        ),
    )
    metadata = ma.fields.Dict(
        allow_none=True,
        load_default={},
    )
    architectureType = ma.fields.Str(
        validate=ma.validate.OneOf(
            [model_framework.value for model_framework in ModelType],
            error="`architectureType` must be one of the supported frameworks. "
            + "Check out our API reference for a full list"
            + " https://reference.openlayer.com/reference/api/openlayer.ModelType.html.\n ",
        ),
        required=True,
    )



"""Implements the series of validations needed to ensure that
the objects uploaded to the platform are valid.

Typically, the validator object is created and then the `validate` method is called.
For example, to validate a model package:

>>> model_validator = ModelValidator(
...     model_package_dir="/path/to/model/package",
...     model_config_file_path="/path/to/model/config/file.yaml",
...     sample_data=df)
>>> model_validator.validate()
"""
import ast
import importlib
import os
import warnings
from typing import Any, Dict, List, Optional

import marshmallow as ma
import pandas as pd
import pkg_resources
import yaml

from . import schemas, utils


class BaselineModelValidator:
    """Validates the baseline model.

    Parameters
    ----------
    model_config_file_path : Optional[str], optional
        The path to the model config file, by default None
    """

    def __init__(self, model_config_file_path: Optional[str] = None):
        self.model_config_file_path = model_config_file_path
        self.failed_validations = []

    def _validate_model_config(self):
        """Validates the model config file."""
        model_config_failed_validations = []

        # File existence check
        if self.model_config_file_path:
            if not os.path.isfile(os.path.expanduser(self.model_config_file_path)):
                model_config_failed_validations.append(
                    f"File `{self.model_config_file_path}` does not exist."
                )
            else:
                with open(self.model_config_file_path, "r", encoding="UTF-8") as stream:
                    model_config = yaml.safe_load(stream)

        if model_config:
            baseline_model_schema = schemas.BaselineModelSchema()
            try:
                baseline_model_schema.load(model_config)
            except ma.ValidationError as err:
                model_config_failed_validations.extend(
                    _format_marshmallow_error_message(err)
                )

        # Print results of the validation
        if model_config_failed_validations:
            print("Baseline model config failed validations: \n")
            _list_failed_validation_messages(model_config_failed_validations)

        # Add the `model_config.yaml` failed validations to the list of all failed validations
        self.failed_validations.extend(model_config_failed_validations)

    def validate(self) -> List[str]:
        """Validates the baseline model.

        Returns
        -------
        List[str]
            The list of failed validations.
        """
        if self.model_config_file_path:
            self._validate_model_config()

        if not self.failed_validations:
            print("All baseline model validations passed!")

        return self.failed_validations


class CommitBundleValidator:
    """Validates the commit bundle prior to push.

    Parameters
    ----------
    bundle_path : str
        The path to the commit bundle (staging area, if for the Python API).
    skip_model_validation : bool
        Whether to skip model validation, by default False
    skip_dataset_validation : bool
        Whether to skip dataset validation, by default False
    """

    def __init__(
        self,
        bundle_path: str,
        skip_model_validation: bool = False,
        skip_dataset_validation: bool = False,
    ):
        self.bundle_path = bundle_path
        self._bundle_resources = utils.list_resources_in_bundle(bundle_path)
        self._skip_model_validation = skip_model_validation
        self._skip_dataset_validation = skip_dataset_validation
        self.failed_validations = []

    def _validate_bundle_state(self):
        """Checks whether the bundle is in a valid state.

        This includes:
        - When a "model" is included, you always need to provide predictions for both
          "validation" and "training" (regardless of artifact or no artifact).
        - When a "baseline-model" is included, you always need to provide a "training"
          and "validation" set without predictions.
        - When a "model" nor a "baseline-model" are included, you always need to NOT
          upload predictions with one exception:
            - "validation" set only in bundle, which means the predictions are for the
            previous model version.
        """
        bundle_state_failed_validations = []

        # Defining which datasets contain predictions
        training_predictions_column_name = None
        validation_predictions_column_name = None
        if "training" in self._bundle_resources:
            with open(
                f"{self.bundle_path}/training/dataset_config.yaml",
                "r",
                encoding="UTF-8",
            ) as stream:
                training_dataset_config = yaml.safe_load(stream)

            training_predictions_column_name = training_dataset_config.get(
                "predictionsColumnName"
            )

        if "validation" in self._bundle_resources:
            with open(
                f"{self.bundle_path}/validation/dataset_config.yaml",
                "r",
                encoding="UTF-8",
            ) as stream:
                validation_dataset_config = yaml.safe_load(stream)

            validation_predictions_column_name = validation_dataset_config.get(
                "predictionsColumnName"
            )

        if "model" in self._bundle_resources:
            if (
                training_predictions_column_name is None
                or validation_predictions_column_name is None
            ):
                bundle_state_failed_validations.append(
                    "To push a model to the platform, you must provide "
                    "training and a validation sets with predictions in the column "
                    "`predictions_column_name`."
                )
        elif "baseline-model" in self._bundle_resources:
            if (
                "training" not in self._bundle_resources
                or "validation" not in self._bundle_resources
            ):
                bundle_state_failed_validations.append(
                    "To push a baseline model to the platform, you must provide "
                    "training and validation sets."
                )
            elif (
                training_predictions_column_name is not None
                and validation_predictions_column_name is not None
            ):
                bundle_state_failed_validations.append(
                    "To push a baseline model to the platform, you must not provide "
                    "training and a validation sets without predictions in the column "
                    "`predictions_column_name`."
                )
        else:
            if (
                "training" in self._bundle_resources
                and validation_predictions_column_name is not None
            ):
                bundle_state_failed_validations.append(
                    "A training set was provided alongside with a validation set with"
                    " predictions. Please either provide only a validation set with"
                    " predictions, or a model and both datasets with predictions"
                )
            elif training_predictions_column_name is not None:
                bundle_state_failed_validations.append(
                    "The training dataset contains predictions, but no model was"
                    " provided. To push a training set with predictions, please provide"
                    " a model and a validation set with predictions as well."
                )

        # Print results of the validation
        if bundle_state_failed_validations:
            print("Push failed validations: \n")
            _list_failed_validation_messages(bundle_state_failed_validations)

        # Add the bundle state failed validations to the list of all failed validations
        self.failed_validations.extend(bundle_state_failed_validations)

    def _validate_bundle_resources(self):
        """Runs the corresponding validations for each resource in the bundle."""
        bundle_resources_failed_validations = []

        if "training" in self._bundle_resources and not self._skip_dataset_validation:
            training_set_validator = DatasetValidator(
                dataset_config_file_path=f"{self.bundle_path}/training/dataset_config.yaml",
                dataset_file_path=f"{self.bundle_path}/training/dataset.csv",
            )
            bundle_resources_failed_validations.extend(
                training_set_validator.validate()
            )

        if "validation" in self._bundle_resources and not self._skip_dataset_validation:
            validation_set_validator = DatasetValidator(
                dataset_config_file_path=f"{self.bundle_path}/validation/dataset_config.yaml",
                dataset_file_path=f"{self.bundle_path}/training/dataset.csv",
            )
            bundle_resources_failed_validations.extend(
                validation_set_validator.validate()
            )

        if (
            "baseline-model" in self._bundle_resources
            and not self._skip_model_validation
        ):
            baseline_model_validator = BaselineModelValidator(
                model_config_file_path=f"{self.bundle_path}/baseline-model/model_config.yaml"
            )
            bundle_resources_failed_validations.extend(
                baseline_model_validator.validate()
            )

        if "model" in self._bundle_resources and not self._skip_model_validation:
            model_files = os.listdir(f"{self.bundle_path}/model")
            # Shell model
            if len(model_files) == 1:
                model_validator = ModelValidator(
                    model_config_file_path=f"{self.bundle_path}/model/model_config.yaml"
                )
            # Model package
            else:
                # Use data from the validation as test data
                validation_dataset_df = self._load_dataset_from_bundle("validation")
                validation_dataset_config = self._load_dataset_config_from_bundle(
                    "validation"
                )

                sample_data = None
                if "textColumnName" in validation_dataset_config:
                    sample_data = validation_dataset_df[
                        validation_dataset_config["textColumnName"]
                    ].head()

                else:
                    sample_data = validation_dataset_df[
                        validation_dataset_config["featureNames"]
                    ].head()

                model_validator = ModelValidator(
                    model_config_file_path=f"{self.bundle_path}/model/model_config.yaml",
                    model_package_dir=f"{self.bundle_path}/model",
                    sample_data=sample_data,
                )
                bundle_resources_failed_validations.extend(model_validator.validate())

        # Print results of the validation
        if bundle_resources_failed_validations:
            print("Push failed validations: \n")
            _list_failed_validation_messages(bundle_resources_failed_validations)

        # Add the bundle resources failed validations to the list of all failed validations
        self.failed_validations.extend(bundle_resources_failed_validations)

    def _load_dataset_from_bundle(self, label: str) -> pd.DataFrame:
        """Loads a dataset from a commit bundle.

        Parameters
        ----------
        label : str
            The type of the dataset. Can be either "training" or "validation".

        Returns
        -------
        pd.DataFrame
            The dataset.
        """
        dataset_file_path = f"{self.bundle_path}/{label}/dataset.csv"

        dataset_df = pd.read_csv(dataset_file_path)

        return dataset_df

    def _load_dataset_config_from_bundle(self, label: str) -> Dict[str, Any]:
        """Loads a dataset config from a commit bundle.

        Parameters
        ----------
        label : str
            The type of the dataset. Can be either "training" or "validation".

        Returns
        -------
        Dict[str, Any]
            The dataset config.
        """
        dataset_config_file_path = f"{self.bundle_path}/{label}/dataset_config.yaml"

        with open(dataset_config_file_path, "r", encoding="UTF-8") as stream:
            dataset_config = yaml.safe_load(stream)

        return dataset_config

    def validate(self) -> List[str]:
        """Validates the commit bundle.

        Returns
        -------
        List[str]
            A list of failed validations.
        """
        self._validate_bundle_state()
        self._validate_bundle_resources()

        if not self.failed_validations:
            print("All commit bundle validations passed!")

        return self.failed_validations


class CommitValidator:
    """Validates the commit prior to the upload.

    Parameters
    ----------
    commit_message : str
        The commit message.
    """

    def __init__(
        self,
        commit_message: str,
    ):
        self.commit_message = commit_message
        self.failed_validations = []

    def _validate_commit_message(self):
        """Checks whether the commit message is valid."""
        commit_message_failed_validations = []

        commit_schema = schemas.CommitSchema()
        try:
            commit_schema.load({"commitMessage": self.commit_message})
        except ma.ValidationError as err:
            commit_message_failed_validations.extend(
                _format_marshmallow_error_message(err)
            )

        # Print results of the validation
        if commit_message_failed_validations:
            print("Commit failed validations: \n")
            _list_failed_validation_messages(commit_message_failed_validations)

        # Add the commit failed validations to the list of all failed validations
        self.failed_validations.extend(commit_message_failed_validations)

    def validate(self) -> List[str]:
        """Validates the commit.

        Returns
        -------
        List[str]
            A list of failed validations.
        """
        self._validate_commit_message()

        if not self.failed_validations:
            print("All commit validations passed!")

        return self.failed_validations


class DatasetValidator:
    """Validates the dataset and its arguments.

    Either the ``dataset_file_path`` or the ``dataset_df`` must be
    provided (not both).

    Either the ``dataset_config_file_path`` or the ``dataset_config``
    must be provided (not both).

    Parameters
    ----------
    dataset_config_file_path : str, optional
        The path to the dataset_config.yaml file.
    dataset_config : dict, optional
        The dataset_config as a dictionary.
    dataset_file_path : str, optional
        The path to the dataset file.
    dataset_df : pd.DataFrame, optional
        The dataset to validate.

    Examples
    --------

    Let's say we have a ``dataset_config.yaml`` file and a ``dataset.csv``
    file in the current directory.

    To ensure they are in the format the Openlayer platform expects to use the
    :meth:`openlayer.OpenlayerClient.add_dataset`, we can use the
    :class:`openlayer.DatasetValidator` class as follows:

    >>> from openlayer import DatasetValidator
    >>>
    >>> dataset_validator = DatasetValidator(
    ...     dataset_config_file_path="dataset_config.yaml",
    ...     dataset_file_path="dataset.csv",
    ... )
    >>> dataset_validator.validate()

    Alternatively, if we have a ``dataset_config.yaml`` file in the current
    directory and a ``dataset_df`` DataFrame, we can use the
    :class:`openlayer.DatasetValidator` class as follows:

    >>> from openlayer import DatasetValidator
    >>>
    >>> dataset_validator = DatasetValidator(datas
    ...     dataset_config_file_path="dataset_config.yaml",
    ...     dataset_df=dataset_df,
    ... )
    >>> dataset_validator.validate()
    """

    def __init__(
        self,
        dataset_config_file_path: Optional[str] = None,
        dataset_config: Optional[Dict] = None,
        dataset_file_path: Optional[str] = None,
        dataset_df: Optional[pd.DataFrame] = None,
    ):
        if dataset_df is not None and dataset_file_path:
            raise ValueError(
                "Both dataset_df and dataset_file_path are provided."
                " Please provide only one of them."
            )
        elif dataset_df is None and not dataset_file_path:
            raise ValueError(
                "Neither dataset_df nor dataset_file_path is provided."
                " Please provide one of them."
            )

        if dataset_config_file_path and dataset_config:
            raise ValueError(
                "Both dataset_config_file_path and dataset_config are provided."
                " Please provide only one of them."
            )
        elif not dataset_config_file_path and not dataset_config:
            raise ValueError(
                "Neither dataset_config_file_path nor dataset_config is provided."
                " Please provide one of them."
            )

        self.dataset_file_path = dataset_file_path
        self.dataset_df = dataset_df
        self.dataset_config_file_path = dataset_config_file_path
        self.dataset_config = dataset_config
        self.failed_validations = []

    def _validate_dataset_config(self):
        """Checks whether the dataset_config is valid.

        Beware of the order of the validations, as it is important.
        """
        dataset_config_failed_validations = []

        # File existence check
        if self.dataset_config_file_path:
            if not os.path.isfile(os.path.expanduser(self.dataset_config_file_path)):
                dataset_config_failed_validations.append(
                    f"File `{self.dataset_config_file_path}` does not exist."
                )
            else:
                with open(
                    self.dataset_config_file_path, "r", encoding="UTF-8"
                ) as stream:
                    self.dataset_config = yaml.safe_load(stream)

        if self.dataset_config:
            dataset_schema = schemas.DatasetSchema()
            try:
                dataset_schema.load(self.dataset_config)
            except ma.ValidationError as err:
                dataset_config_failed_validations.extend(
                    _format_marshmallow_error_message(err)
                )

        # Print results of the validation
        if dataset_config_failed_validations:
            print("Dataset_config failed validations: \n")
            _list_failed_validation_messages(dataset_config_failed_validations)

        # Add the `dataset_config.yaml` failed validations to the list of all failed validations
        self.failed_validations.extend(dataset_config_failed_validations)

    def _validate_dataset_file(self):
        """Checks whether the dataset file exists and is valid.

        If it is valid, it loads the dataset file into the `self.dataset_df`
        attribute.

        Beware of the order of the validations, as it is important.
        """
        dataset_file_failed_validations = []

        # File existence check
        if not os.path.isfile(os.path.expanduser(self.dataset_file_path)):
            dataset_file_failed_validations.append(
                f"File `{self.dataset_file_path}` does not exist."
            )
        else:
            # File format (csv) check by loading it as a pandas df
            try:
                self.dataset_df = pd.read_csv(self.dataset_file_path)
            except Exception:
                dataset_file_failed_validations.append(
                    f"File `{self.dataset_file_path}` is not a valid .csv file."
                )

        # Print results of the validation
        if dataset_file_failed_validations:
            print("Dataset file failed validations: \n")
            _list_failed_validation_messages(dataset_file_failed_validations)

        # Add the dataset file failed validations to the list of all failed validations
        self.failed_validations.extend(dataset_file_failed_validations)

    def _validate_dataset_and_config_consistency(self):
        """Checks whether the dataset and its config are consistent.

        Beware of the order of the validations, as it is important.
        """
        dataset_and_config_consistency_failed_validations = []

        if self.dataset_config and self.dataset_df is not None:
            # Extract vars
            dataset_df = self.dataset_df
            categorical_feature_names = self.dataset_config.get(
                "categoricalFeatureNames"
            )
            class_names = self.dataset_config.get("classNames")
            column_names = self.dataset_config.get("columnNames")
            label_column_name = self.dataset_config.get("labelColumnName")
            feature_names = self.dataset_config.get("featureNames")
            text_column_name = self.dataset_config.get("textColumnName")
            predictions_column_name = self.dataset_config.get("predictionsColumnName")

            if self._contains_unsupported_dtypes(dataset_df):
                dataset_and_config_consistency_failed_validations.append(
                    "The dataset contains unsupported dtypes. The supported dtypes are "
                    "'float32', 'float64', 'int32', 'int64', 'object'."
                    " Please cast the columns in your dataset to conform to these dtypes."
                )

            if self._columns_not_specified(dataset_df, column_names):
                dataset_and_config_consistency_failed_validations.append(
                    "Not all columns in the dataset are specified in `columnNames`."
                    " Please specify all dataset columns in `columnNames`."
                )

            if label_column_name:
                if self._column_not_in_dataset_df(dataset_df, label_column_name):
                    dataset_and_config_consistency_failed_validations.append(
                        f"The label column `{label_column_name}` specified as `labelColumnName` "
                        "is not in the dataset."
                    )
                else:
                    if class_names:
                        if self._labels_not_in_class_names(
                            dataset_df, label_column_name, class_names
                        ):
                            dataset_and_config_consistency_failed_validations.append(
                                "There are more labels in the dataset's column"
                                f" `{label_column_name}` than specified in `classNames`. "
                                "Please specify all possible labels in the `classNames` list."
                            )
                        if self._labels_not_zero_indexed(
                            dataset_df, label_column_name, class_names
                        ):
                            dataset_and_config_consistency_failed_validations.append(
                                "The labels in the dataset are not zero-indexed. "
                                f"Make sure that the labels in the column `{label_column_name}` "
                                "are zero-indexed integers that match the list in `classNames`."
                            )

            # Predictions validations
            if predictions_column_name:
                if self._column_not_in_dataset_df(dataset_df, predictions_column_name):
                    dataset_and_config_consistency_failed_validations.append(
                        f"The predictions column `{predictions_column_name}` specified as"
                        " `predictionsColumnName` is not in the dataset."
                    )
                else:
                    try:
                        # Getting prediction lists from strings saved in the csv
                        dataset_df[predictions_column_name] = dataset_df[
                            predictions_column_name
                        ].apply(ast.literal_eval)
                        if self._predictions_not_lists(
                            dataset_df, predictions_column_name
                        ):
                            dataset_and_config_consistency_failed_validations.append(
                                f"The predictions in the column `{predictions_column_name}` "
                                "are not lists. Please make sure that the predictions are "
                                "lists of floats."
                            )
                        else:
                            if self._prediction_lists_not_same_length(
                                dataset_df, predictions_column_name
                            ):
                                dataset_and_config_consistency_failed_validations.append(
                                    "The prediction lists in the column "
                                    f"`{predictions_column_name}` "
                                    "are not all of the same length. "
                                    "Please make sure that all prediction lists "
                                    "are of the same length."
                                )
                            else:
                                if self._predictions_not_class_probabilities(
                                    dataset_df, predictions_column_name
                                ):
                                    dataset_and_config_consistency_failed_validations.append(
                                        "The predictions in the column "
                                        f"`{predictions_column_name}` "
                                        "are not class probabilities. "
                                        "Please make sure that the predictions are lists "
                                        "of floats that sum to 1."
                                    )
                                elif class_names:
                                    if self._predictions_not_in_class_names(
                                        dataset_df, predictions_column_name, class_names
                                    ):
                                        dataset_and_config_consistency_failed_validations.append(
                                            "There are predictions in the column "
                                            f"`{predictions_column_name}` "
                                            "are not in `classNames`. "
                                            "Please make sure that the predictions are lists "
                                            "of floats that sum to 1 and that the classes in "
                                            "the predictions match the classes in `classNames`."
                                        )
                    except Exception:
                        dataset_and_config_consistency_failed_validations.append(
                            f"The predictions in the column `{predictions_column_name}` "
                            "are not lists. "
                            "Please make sure that the predictions are lists of floats."
                        )

            # NLP-specific validations
            if text_column_name:
                if self._column_not_in_dataset_df(dataset_df, text_column_name):
                    dataset_and_config_consistency_failed_validations.append(
                        f"The text column `{text_column_name}` specified as `textColumnName` "
                        "is not in the dataset."
                    )
                elif self._exceeds_character_limit(dataset_df, text_column_name):
                    dataset_and_config_consistency_failed_validations.append(
                        f"The column `{text_column_name}` of the dataset contains rows that "
                        "exceed the 1000 character limit."
                    )

            # Tabular-specific validations
            if feature_names:
                if self._columns_not_in_dataset_df(dataset_df, feature_names):
                    dataset_and_config_consistency_failed_validations.append(
                        "There are features specified in `featureNames` which are "
                        "not in the dataset."
                    )
                if categorical_feature_names:
                    if self._columns_not_in_dataset_df(
                        dataset_df, categorical_feature_names
                    ):
                        dataset_and_config_consistency_failed_validations.append(
                            "There are categorical features specified in `categoricalFeatureNames` "
                            "which are not in the dataset."
                        )

        # Print results of the validation
        if dataset_and_config_consistency_failed_validations:
            print("Inconsistencies between the dataset config and the dataset: \n")
            _list_failed_validation_messages(
                dataset_and_config_consistency_failed_validations
            )

        # Add the consistency failed validations to the list of all failed validations
        self.failed_validations.extend(
            dataset_and_config_consistency_failed_validations
        )

    @staticmethod
    def _labels_not_zero_indexed(
        dataset_df: pd.DataFrame, label_column_name: str, class_names: List[str]
    ) -> bool:
        """Checks whether the labels are zero-indexed."""
        unique_labels = set(dataset_df[label_column_name].unique())
        zero_indexed_set = set(range(len(class_names)))
        if unique_labels != zero_indexed_set:
            return True
        return False

    @staticmethod
    def _contains_unsupported_dtypes(dataset_df: pd.DataFrame) -> bool:
        """Checks whether the dataset contains unsupported dtypes."""
        supported_dtypes = {"float32", "float64", "int32", "int64", "object"}
        dataset_df_dtypes = set([dtype.name for dtype in dataset_df.dtypes])
        unsupported_dtypes = dataset_df_dtypes - supported_dtypes
        if unsupported_dtypes:
            return True
        return False

    @staticmethod
    def _column_not_in_dataset_df(dataset_df: pd.DataFrame, column_name: str) -> bool:
        """Checks whether the label column is in the dataset."""
        if column_name not in dataset_df.columns:
            return True
        return False

    @staticmethod
    def _columns_not_in_dataset_df(
        dataset_df: pd.DataFrame, columns_list: List[str]
    ) -> bool:
        """Checks whether the columns are in the dataset."""
        if set(columns_list) - set(dataset_df.columns):
            return True
        return False

    @staticmethod
    def _columns_not_specified(
        dataset_df: pd.DataFrame, columns_list: List[str]
    ) -> bool:
        """Checks whether the columns are specified."""
        if set(columns_list) != set(dataset_df.columns):
            return True
        return False

    @staticmethod
    def _exceeds_character_limit(
        dataset_df: pd.DataFrame, text_column_name: str
    ) -> bool:
        """Checks whether the text column exceeds the character limit."""
        if dataset_df[text_column_name].str.len().max() > 1000:
            return True
        return False

    @staticmethod
    def _labels_not_in_class_names(
        dataset_df: pd.DataFrame, label_column_name: str, class_names: List[str]
    ) -> bool:
        """Checks whether there are labels in the dataset which are not
        in the `class_names`."""
        num_classes = len(dataset_df[label_column_name].unique())
        if num_classes > len(class_names):
            return True
        return False

    @staticmethod
    def _predictions_not_lists(
        dataset_df: pd.DataFrame, predictions_column_name: str
    ) -> bool:
        """Checks whether all values in the column `predictions_column_name`
        are lists."""
        if not all(
            isinstance(predictions, list)
            for predictions in dataset_df[predictions_column_name]
        ):
            return True
        return False

    @staticmethod
    def _prediction_lists_not_same_length(
        dataset_df: pd.DataFrame, predictions_column_name: str
    ) -> bool:
        """Checks whether all the lists in the `predictions_column_name`
        have the same length."""
        if not len(set(dataset_df[predictions_column_name].str.len())) == 1:
            return True
        return False

    @staticmethod
    def _predictions_not_class_probabilities(
        dataset_df: pd.DataFrame, predictions_column_name: str
    ) -> bool:
        """Checks whether the predictions are class probabilities.
        Tolerate a 10% error margin."""
        if any(
            [
                (sum(predictions) < 0.9 or sum(predictions) > 1.1)
                for predictions in dataset_df[predictions_column_name]
            ]
        ):
            return True
        return False

    @staticmethod
    def _predictions_not_in_class_names(
        dataset_df: pd.DataFrame,
        predictions_column_name: str,
        class_names: List[str],
    ) -> bool:
        """Checks if the predictions map 1:1 to the `class_names` list."""
        num_classes_predicted = len(dataset_df[predictions_column_name].iloc[0])
        if num_classes_predicted != len(class_names):
            return True
        return False

    def validate(self) -> List[str]:
        """Runs all dataset validations.

        At each stage, prints all the failed validations.

        Returns
        -------
        List[str]
            List of all failed validations.
        """
        self._validate_dataset_config()
        if self.dataset_file_path:
            self._validate_dataset_file()
        self._validate_dataset_and_config_consistency()

        if not self.failed_validations:
            print("All dataset validations passed!")

        return self.failed_validations


class ModelValidator:
    """Validates the model package's structure and files.

    Parameters
    ----------
    model_package_dir : str
        Path to the model package directory.
    sample_data : pd.DataFrame
        Sample data to be used for the model validation.

    Examples
    --------

    Let's say we have the prepared the model package and have some sample data expected
    by the model in a pandas DataFrame.

    To ensure the model package is in the format the Openlayer platform expects to use the
    :meth:`openlayer.OpenlayerClient.add_model` method, we can use the
    :class:`openlayer.ModelValidator` class as follows:,

    >>> from openlayer import ModelValidator
    >>>
    >>> model_validator = ModelValidator(
    ...     model_package_dir="/path/to/model/package",
    ...     sample_data=df,
    ... )
    >>> model_validator.validate()

    """

    def __init__(
        self,
        model_config_file_path: str,
        model_package_dir: Optional[str] = None,
        sample_data: Optional[pd.DataFrame] = None,
    ):
        self.model_config_file_path = model_config_file_path
        self.model_package_dir = model_package_dir
        self.sample_data = sample_data
        self.failed_validations = []

    def _validate_model_package_dir(self):
        """Verifies the model package directory structure.

        The model package directory must follow the structure:

        model_package
          ├── artifacts.pkl  # potentially different name / format and multiple files
          ├── prediction_interface.py
          └── requirements.txt

        This method checks for the existence of the above files.
        """
        model_package_failed_validations = []

        if not os.path.exists(self.model_package_dir):
            model_package_failed_validations.append(
                f"Model package directory `{self.model_package_dir}` does not exist."
            )

        if not os.path.isdir(self.model_package_dir):
            model_package_failed_validations.append(
                f"Model package directory `{self.model_package_dir}` is not a directory."
            )

        if self.model_package_dir == os.getcwd():
            model_package_failed_validations.append(
                f"Model package directory `{self.model_package_dir}` is the current "
                "working directory."
            )

        if not os.path.exists(
            os.path.join(self.model_package_dir, "prediction_interface.py")
        ):
            model_package_failed_validations.append(
                f"Model package directory `{self.model_package_dir}` does not contain the "
                "`prediction_interface.py` file."
            )

        if not os.path.exists(os.path.join(self.model_package_dir, "requirements.txt")):
            model_package_failed_validations.append(
                f"Model package directory `{self.model_package_dir}` does not contain the "
                "`requirements.txt` file."
            )

        # Print results of the validation
        if model_package_failed_validations:
            print("Model package structure failed validations: \n")
            _list_failed_validation_messages(model_package_failed_validations)

        # Add the model package failed validations to the list of all failed validations
        self.failed_validations.extend(model_package_failed_validations)

    def _validate_requirements(self):
        """Validates the requirements.txt file.

        Checks for the existence of the file and parses it to check for
        version discrepancies. Appends to the list of failed validations,
        if the file does not exist, and raises warnings in case of
        discrepancies.

        Beware of the order of the validations, as it is important.
        """
        requirements_failed_validations = []

        # Path to the requirements.txt file
        requirements_txt_file = os.path.join(self.model_package_dir, "requirements.txt")

        # File existence check
        if not os.path.isfile(os.path.expanduser(requirements_txt_file)):
            requirements_failed_validations.append(
                f"File `{requirements_txt_file}` does not exist."
            )
        else:
            with open(requirements_txt_file, "r", encoding="UTF-8") as file:
                lines = file.readlines()

            # Parse the requirements file
            requirements = pkg_resources.parse_requirements(lines)

            for requirement in requirements:
                requirement = str(requirement)

                # Consistency checks between requirements and modules installed in the environment
                try:
                    pkg_resources.require(requirement)
                except pkg_resources.VersionConflict as err:
                    try:
                        warnings.warn(
                            "There is a version discrepancy between the current "
                            f"environment and the dependency `{requirement}`. \n"
                            f"`requirements.txt` specifies `{err.req}`, but the current "
                            f"environment contains `{err.dist}` installed. \n"
                            "There might be unexpected results once the model is in the platform. "
                            "Use at your own discretion.",
                            category=Warning,
                        )
                        return None
                    except AttributeError:
                        warnings.warn(
                            "There is a version discrepancy between the current "
                            f"environment and the dependency `{requirement}`. \n"
                            f"`requirements.txt` specifies `{requirement}`, but the current "
                            f"environment contains an incompatible version installed. \n"
                            "There might be unexpected results once the model is in the platform. "
                            "Use at your own discretion.",
                            category=Warning,
                        )
                        return None
                except pkg_resources.DistributionNotFound:
                    warnings.warn(
                        f"The dependency `{requirement}` specified in the `requirements.txt` "
                        "is not installed in the current environment. \n"
                        "There might be unexpected results once the model is in the platform. "
                        "Use at your own discretion.",
                        category=Warning,
                    )

        # Print results of the validation
        if requirements_failed_validations:
            print("`requirements.txt` failed validations: \n")
            _list_failed_validation_messages(requirements_failed_validations)

        # Add the `requirements.txt` failed validations to the list of all failed validations
        self.failed_validations.extend(requirements_failed_validations)

    def _validate_model_config(self):
        """Checks whether the model_config.yaml file exists and is valid.

        Beware of the order of the validations, as it is important.
        """
        model_config_failed_validations = []

        # File existence check
        if not os.path.isfile(os.path.expanduser(self.model_config_file_path)):
            model_config_failed_validations.append(
                f"File `{self.model_config_file_path}` does not exist."
            )
        else:
            with open(self.model_config_file_path, "r", encoding="UTF-8") as stream:
                model_config = yaml.safe_load(stream)

            if self.model_package_dir:
                model_schema = schemas.ModelSchema()
            else:
                model_schema = schemas.ShellModelSchema()
            try:
                model_schema.load(model_config)
            except ma.ValidationError as err:
                model_config_failed_validations.extend(
                    _format_marshmallow_error_message(err)
                )

        # Print results of the validation
        if model_config_failed_validations:
            print("`model_config.yaml` failed validations: \n")
            _list_failed_validation_messages(model_config_failed_validations)

        # Add the `model_config.yaml` failed validations to the list of all failed validations
        self.failed_validations.extend(model_config_failed_validations)

    def _validate_prediction_interface(self):
        """Validates the implementation of the prediction interface.

        Checks for the existence of the file, the required functions, and
        runs test data through the model to ensure there are no implementation
        errors.

        Beware of the order of the validations, as it is important.
        """
        prediction_interface_failed_validations = []

        # Path to the prediction_interface.py file
        prediction_interface_file = os.path.join(
            self.model_package_dir, "prediction_interface.py"
        )

        # File existence check
        if not os.path.isfile(os.path.expanduser(prediction_interface_file)):
            prediction_interface_failed_validations.append(
                f"File `{prediction_interface_file}` does not exist."
            )
        else:
            # Loading the module defined in the prediction_interface.py file
            module_spec = importlib.util.spec_from_file_location(
                "model_module", prediction_interface_file
            )
            module = importlib.util.module_from_spec(module_spec)
            module_spec.loader.exec_module(module)

            # Check if the module contains the required functions
            if not hasattr(module, "load_model"):
                prediction_interface_failed_validations.append(
                    "The `load_model` function is not defined in the `prediction_interface.py` "
                    "file."
                )
            else:
                # Test `load_model` function
                ml_model = None
                try:
                    ml_model = module.load_model()
                except Exception as exc:
                    prediction_interface_failed_validations.append(
                        f"There is an error while loading the model: \n {exc}"
                    )

                if ml_model is not None:
                    # Check if the `predict_proba` method is part of the model object
                    if not hasattr(ml_model, "predict_proba"):
                        prediction_interface_failed_validations.append(
                            "The `predict_proba` function is not defined in the model class."
                        )
                    else:
                        # Test `predict_proba` function
                        try:
                            with utils.HidePrints():
                                ml_model.predict_proba(self.sample_data)
                        except Exception as exc:
                            exception_stack = utils.get_exception_stacktrace(exc)
                            prediction_interface_failed_validations.append(
                                "The `predict_proba` function failed while running the test data. "
                                "It is failing with the following error message: \n"
                                f"{exception_stack}"
                            )

        # Print results of the validation
        if prediction_interface_failed_validations:
            print("`prediction_interface.py` failed validations: \n")
            _list_failed_validation_messages(prediction_interface_failed_validations)

        # Add the `prediction_interface.py` failed validations to the list of all failed validations
        self.failed_validations.extend(prediction_interface_failed_validations)

    def validate(self) -> List[str]:
        """Runs all model validations.

        At each stage, prints all the failed validations.

        Returns
        -------
        List[str]
            A list of all failed validations.
        """
        if self.model_package_dir:
            self._validate_model_package_dir()
            self._validate_requirements()
            self._validate_prediction_interface()
        self._validate_model_config()

        if not self.failed_validations:
            print("All validations passed!")

        return self.failed_validations


class ProjectValidator:
    """Validates the project.

    Parameters
    ----------
    project_config : Dict[str, str]
        The project configuration.
    """

    def __init__(
        self,
        project_config: Dict[str, str],
    ):
        self.project_config = project_config
        self.failed_validations = []

    def _validate_project_config(self):
        """Checks if the project configuration is valid."""
        project_config_failed_validations = []

        project_schema = schemas.ProjectSchema()
        try:
            project_schema.load(
                {
                    "name": self.project_config.get("name"),
                    "description": self.project_config.get("description"),
                    "task_type": self.project_config.get("task_type").value,
                }
            )
        except ma.ValidationError as err:
            project_config_failed_validations.extend(
                _format_marshmallow_error_message(err)
            )

        # Print results of the validation
        if project_config_failed_validations:
            print("Project config failed validations: \n")
            _list_failed_validation_messages(project_config_failed_validations)

        # Add the commit failed validations to the list of all failed validations
        self.failed_validations.extend(project_config_failed_validations)

    def validate(self):
        """Validates the project."""
        self._validate_project_config()

        if not self.failed_validations:
            print("All model validations passed!")

        return self.failed_validations


# ----------------------------- Helper functions ----------------------------- #
def _format_marshmallow_error_message(err: ma.ValidationError) -> List[str]:
    """Formats the error messages from Marshmallow to conform to the expected
    list of strings format.

    Parameters
    ----------
    err : ma.ValidationError
        The error object returned by Marshmallow.

    Returns
    -------
    List[str]
        A list of strings, where each string is a failed validation.
    """
    error_msg = []
    for input_data, msg in err.messages.items():
        if input_data == "_schema":
            temp_msg = "\n".join(msg)
            error_msg.append(f"{temp_msg}")
        elif not isinstance(msg, dict):
            temp_msg = msg[0].lower()
            error_msg.append(f"`{input_data}`: {temp_msg}")
        else:
            temp_msg = list(msg.values())[0][0].lower()
            error_msg.append(f"`{input_data}` contains items that are {temp_msg}")

    return error_msg


def _list_failed_validation_messages(failed_validations: List[str]):
    """Prints the failed validations in a list format, with one failed
    validation per line."""
    for msg in failed_validations:
        print(f"- {msg} \n")



"""A collection of the different Openlayer Python client exceptions and their error codes.

Typical usage example:

  if project is None:
    raise errors.OpenlayerResourceNotFound(f"Project {project_id} does not exist")
"""
from typing import Dict


class OpenlayerException(Exception):
    """Generic OpenlayerException class"""

    code = None

    def __init__(self, message, errcode=None):
        if not message:
            message = type(self).__name__
        self.message = message

        if errcode:
            self.code = errcode

        if self.code:
            super().__init__(f"<Response [{self.code}]> {message}")
        else:
            super().__init__(f"<Response> {message}")


class OpenlayerValidationError(OpenlayerException):
    """Failed resource validations"""

    def __init__(self, message):
        super().__init__(message)


class OpenlayerSubscriptionPlanException(OpenlayerException):
    """Subscription plan exception class"""

    def __init__(self, message, context=None, mitigation=None):
        context = context or "You have reached your subscription plan's limits. \n"
        mitigation = mitigation or "To upgrade your plan, visit https://openlayer.com"
        super().__init__(context + message + mitigation)


class OpenlayerInvalidRequest(OpenlayerException):
    """400 - Bad Request -- The request was unacceptable,
    often due to missing a required parameter.
    """

    code = 400


class OpenlayerUnauthorized(OpenlayerException):
    """401 - Unauthorized -- No valid API key provided."""

    code = 401


class OpenlayerNotEnabled(OpenlayerException):
    """402 - Not enabled -- Please contact sales@openlayer.com before
    creating this type of task.
    """

    code = 402


class OpenlayerResourceNotFound(OpenlayerException):
    """404 - Not Found -- The requested resource doesn't exist."""

    code = 404


class OpenlayerDuplicateTask(OpenlayerException):
    """409 - Conflict -- The provided idempotency key or unique_id is
    already in use for a different request.
    """

    code = 409


class OpenlayerTooManyRequests(OpenlayerException):
    """429 - Too Many Requests -- Too many requests hit the API
    too quickly.
    """

    code = 429


class OpenlayerInternalError(OpenlayerException):
    """500 - Internal Server Error -- We had a problem with our server.
    Try again later.
    """

    code = 500


class OpenlayerServiceUnavailable(OpenlayerException):
    """503 - Server Timeout From Request Queueing -- Try again later."""

    code = 503


class OpenlayerTimeoutError(OpenlayerException):
    """504 - Server Timeout Error -- Try again later."""

    code = 504


ExceptionMap: Dict[int, OpenlayerException] = {
    OpenlayerInvalidRequest.code: OpenlayerInvalidRequest,
    OpenlayerUnauthorized.code: OpenlayerUnauthorized,
    OpenlayerNotEnabled.code: OpenlayerNotEnabled,
    OpenlayerResourceNotFound.code: OpenlayerResourceNotFound,
    OpenlayerDuplicateTask.code: OpenlayerDuplicateTask,
    OpenlayerTooManyRequests.code: OpenlayerTooManyRequests,
    OpenlayerInternalError.code: OpenlayerInternalError,
    OpenlayerTimeoutError.code: OpenlayerTimeoutError,
    OpenlayerServiceUnavailable.code: OpenlayerServiceUnavailable,
}



import os
import shutil
from enum import Enum

import requests
from requests.adapters import HTTPAdapter, Response, Retry
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from .exceptions import ExceptionMap, OpenlayerException
from .version import __version__

# Parameters for HTTP retry
HTTP_TOTAL_RETRIES = 3  # Number of total retries
HTTP_RETRY_BACKOFF_FACTOR = 2  # Wait 1, 2, 4 seconds between retries
HTTP_STATUS_FORCE_LIST = [408, 429] + list(range(500, 504)) + list(range(506, 531))
HTTP_RETRY_ALLOWED_METHODS = frozenset({"GET", "PUT", "POST"})

CLIENT_METADATA = {"version": __version__}


class StorageType(Enum):
    """Storage options for uploads."""

    ONPREM = "local"
    AWS = "s3"
    GCP = "gcs"
    AZURE = "azure"


STORAGE = StorageType.AWS
OPENLAYER_ENDPOINT = "https://api.openlayer.com/v1"


class Api:
    """Internal class to handle http requests"""

    def __init__(self, api_key: str):
        if api_key == "" or api_key is None:
            raise OpenlayerException(
                "There is an issue instantiating the OpenlayerClient. \n"
                "An invalid API key is being provided. \n"
                "Make sure to provide a valid API key using the syntax "
                "`OpenlayerClient('YOUR_API_KEY_HERE')`. You can find your API keys "
                "in the Profile page on the Openlayer platform."
            )

        self.api_key = api_key
        self._headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        }
        self._headers_multipart_form_data = {"Authorization": f"Bearer {self.api_key}"}

    @staticmethod
    def _http_request(
        method,
        url,
        headers=None,
        params=None,
        body=None,
        files=None,
        data=None,
    ) -> Response:
        with requests.Session() as https:
            retry_strategy = Retry(
                total=HTTP_TOTAL_RETRIES,
                backoff_factor=HTTP_RETRY_BACKOFF_FACTOR,
                status_forcelist=HTTP_STATUS_FORCE_LIST,
                method_whitelist=HTTP_RETRY_ALLOWED_METHODS,
                raise_on_status=False,
            )

            adapter = HTTPAdapter(max_retries=retry_strategy)
            https.mount("https://", adapter)

            try:
                params = params or {}
                params.update(CLIENT_METADATA)
                res = https.request(
                    method=method,
                    url=url,
                    headers=headers,
                    params=params,
                    json=body,
                    files=files,
                    data=data,
                )

                return res
            except Exception as err:
                raise OpenlayerException(err) from err

    @staticmethod
    def _raise_on_respose(res: Response):

        try:
            message = res.json().get("error", res.text)
        except ValueError:
            message = res.text

        exception = ExceptionMap.get(res.status_code, OpenlayerException)
        raise exception(message, res.status_code)

    def _api_request(
        self,
        method,
        endpoint,
        headers=None,
        params=None,
        body=None,
        files=None,
        data=None,
    ):
        """Make any HTTP request + error handling."""

        url = f"{OPENLAYER_ENDPOINT}/{endpoint}"

        res = self._http_request(method, url, headers, params, body, files, data)

        json = None
        if res.ok:
            json = res.json()
        else:
            self._raise_on_respose(res)

        return json

    def get_request(self, endpoint: str, params=None):
        """Generic GET Request Wrapper."""
        return self._api_request("GET", endpoint, headers=self._headers, params=params)

    def post_request(self, endpoint: str, body=None, files=None, data=None):
        """Generic POST Request Wrapper."""
        return self._api_request(
            "POST",
            endpoint,
            headers=self._headers
            if files is None
            else self._headers_multipart_form_data,
            body=body,
            files=files,
            data=data,
        )

    def put_request(self, endpoint: str, body=None, files=None, data=None):
        """Generic PUT Request Wrapper."""
        return self._api_request(
            "PUT",
            endpoint,
            headers=self._headers
            if files is None
            else self._headers_multipart_form_data,
            body=body,
            files=files,
            data=data,
        )

    def upload(self, endpoint: str, file_path: str, object_name: str = None, body=None):
        """Generic method to upload data to the default storage medium and create the
        appropriate resource in the backend.
        """
        if STORAGE == StorageType.AWS:
            upload = self.upload_blob_s3
        elif STORAGE == StorageType.GCP:
            upload = self.upload_blob_gcs
        elif STORAGE == StorageType.AZURE:
            upload = self.upload_blob_azure
        else:
            upload = self.transfer_blob
        return upload(
            endpoint=endpoint,
            file_path=file_path,
            object_name=object_name,
            body=body,
        )

    def upload_blob_s3(
        self, endpoint: str, file_path: str, object_name: str = None, body=None
    ):
        """Generic method to upload data to S3 storage and create the appropriate resource
        in the backend.
        """
        params = {"storageInterface": "s3", "objectName": object_name}
        presigned_json = self.get_request(f"{endpoint}/presigned-url", params=params)

        with tqdm(
            total=os.stat(file_path).st_size,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            colour="BLUE",
        ) as t:
            with open(file_path, "rb") as f:
                # Avoid logging here as it will break the progress bar
                fields = presigned_json["fields"]
                fields["file"] = (presigned_json["id"], f)
                e = MultipartEncoder(fields=fields)
                m = MultipartEncoderMonitor(
                    e, lambda monitor: t.update(min(t.total, monitor.bytes_read) - t.n)
                )
                headers = {"Content-Type": m.content_type}
                res = requests.post(presigned_json["url"], data=m, headers=headers)

        if res.ok:
            body["storageUri"] = presigned_json["storageUri"]
            return self.post_request(f"{endpoint}", body=body)
        else:
            self._raise_on_respose(res)

    def upload_blob_gcs(
        self, endpoint: str, file_path: str, object_name: str = None, body=None
    ):
        """Generic method to upload data to Google Cloud Storage and create the
        appropriate resource in the backend.
        """
        params = {"storageInterface": "gcs", "objectName": object_name}
        presigned_json = self.get_request(f"{endpoint}/presigned-url", params=params)
        with open(file_path, "rb") as f:
            with tqdm(
                total=os.stat(file_path).st_size,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            ) as t:
                wrapped_file = CallbackIOWrapper(t.update, f, "read")
                res = requests.put(
                    presigned_json["url"],
                    data=wrapped_file,
                    headers={"Content-Type": "application/x-gzip"},
                )
        if res.ok:
            body["storageUri"] = presigned_json["storageUri"]
            return self.post_request(f"{endpoint}", body=body)
        else:
            self._raise_on_respose(res)

    def upload_blob_azure(
        self, endpoint: str, file_path: str, object_name: str = None, body=None
    ):
        """Generic method to upload data to Azure Blob Storage and create the
        appropriate resource in the backend.
        """
        params = {"storageInterface": "azure", "objectName": object_name}
        presigned_json = self.get_request(f"{endpoint}/presigned-url", params=params)
        with open(file_path, "rb") as f:
            with tqdm(
                total=os.stat(file_path).st_size,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            ) as t:
                wrapped_file = CallbackIOWrapper(t.update, f, "read")
                res = requests.put(
                    presigned_json["url"],
                    data=wrapped_file,
                    headers={
                        "Content-Type": "application/x-gzip",
                        "x-ms-blob-type": "BlockBlob",
                    },
                )
        if res.ok:
            body["storageUri"] = presigned_json["storageUri"]
            return self.post_request(f"{endpoint}", body=body)
        else:
            self._raise_on_respose(res)

    def transfer_blob(self, endpoint: str, file_path: str, object_name: str, body=None):
        """Generic method to transfer data to the openlayer folder and create the appropriate
        resource in the backend when using a local deployment.
        """
        params = {"storageInterface": "local", "objectName": object_name}
        presigned_json = self.get_request(f"{endpoint}/presigned-url", params=params)
        blob_path = presigned_json["storageUri"].replace("local://", "")
        dir_path = os.path.dirname(blob_path)
        try:
            os.makedirs(dir_path, exist_ok=True)
        except OSError:
            raise OpenlayerException(f"Directory {dir_path} cannot be created")
        shutil.copyfile(file_path, blob_path)
        body["storageUri"] = presigned_json["storageUri"]
        return self.post_request(f"{endpoint}", body=body)



"""Series of helper functions and classes that are used throughout the
OpenLayer Python client.
"""
import os
import sys
import traceback
import warnings

import yaml


# -------------------------- Helper context managers ------------------------- #
class HidePrints:
    """Helper class that suppresses the prints and warnings to stdout and Jupyter's stdout.

    Used as a context manager to hide the print / warning statements that can be inside the user's
    function while we test it.
    """

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")
        sys._jupyter_stdout = sys.stdout
        warnings.filterwarnings("ignore")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout
        sys._jupyter_stdout = sys.stdout
        warnings.filterwarnings("default")


# ----------------------------- Helper functions ----------------------------- #
def write_python_version(directory: str):
    """Writes the python version to the file `python_version` in the specified
    directory (`directory`).

    This is used to register the Python version of the user's environment in the
    when they are uploading a model package.

    Args:
        directory (str): the directory to write the file to.
    """
    with open(f"{directory}/python_version", "w", encoding="UTF-8") as file:
        file.write(
            str(sys.version_info.major)
            + "."
            + str(sys.version_info.minor)
            + "."
            + str(sys.version_info.micro)
        )


def remove_python_version(directory: str):
    """Removes the file `python_version` from the specified directory
    (`directory`).

    Args:
        directory (str): the directory to remove the file from.
    """
    os.remove(f"{directory}/python_version")


def read_yaml(filename: str) -> dict:
    """Reads a YAML file and returns it as a dictionary.

    Args:
        filename (str): the path to the YAML file.

    Returns:
        dict: the dictionary representation of the YAML file.
    """
    with open(filename, "r", encoding="UTF-8") as stream:
        return yaml.safe_load(stream)


def write_yaml(dictionary: dict, filename: str):
    """Writes the dictionary to a YAML file in the specified directory (`dir`).

    Args:
        dictionary (dict): the dictionary to write to a YAML file.
        dir (str): the directory to write the file to.
    """
    with open(filename, "w", encoding="UTF-8") as stream:
        yaml.dump(dictionary, stream)


def get_exception_stacktrace(err: Exception):
    """Returns the stacktrace of the most recent exception.

    Returns:
        str: the stacktrace of the most recent exception.
    """
    return "".join(traceback.format_exception(type(err), err, err.__traceback__))


def list_resources_in_bundle(bundle_path: str) -> list:
    """Lists the resources in the bundle.

    Args:
        bundle_path (str): the path to the bundle.

    Returns:
        list: the list of resources in the bundle.
    """
    # TODO: factor out list of valid resources
    VALID_RESOURCES = {"baseline-model", "model", "training", "validation"}

    resources = []

    for resource in os.listdir(bundle_path):
        if resource in VALID_RESOURCES:
            resources.append(resource)
    return resources



"""Module for the Project class.
"""

from . import tasks


class Project:
    """An object containing information about a project on the Openlayer platform."""

    def __init__(self, json, upload, client, subscription_plan=None):
        self._json = json
        self.id = json["id"]
        self.upload = upload
        self.subscription_plan = subscription_plan
        self.client = client

    def __getattr__(self, name):
        if name in self._json:
            return self._json[name]
        raise AttributeError(f"'{type(self).__name__}' object has no attribute {name}")

    def __hash__(self):
        return hash(self.id)

    def __str__(self):
        return f"Project(id={self.id})"

    def __repr__(self):
        return f"Project({self._json})"

    def to_dict(self):
        """Returns object properties as a dict.

        Returns
        -------
        Dict with object properties.
        """
        return self._json

    def add_model(
        self,
        *args,
        **kwargs,
    ):
        """Adds a model to a project's staging area."""
        return self.client.add_model(*args, project_id=self.id, **kwargs)

    def add_baseline_model(
        self,
        *args,
        **kwargs,
    ):
        """Adds a baseline model to the project."""
        return self.client.add_baseline(
            *args, project_id=self.id, task_type=tasks.TaskType(self.taskType), **kwargs
        )

    def add_dataset(
        self,
        *args,
        **kwargs,
    ):
        """Adds a dataset to a project's staging area (from a csv)."""
        return self.client.add_dataset(*args, project_id=self.id, **kwargs)

    def add_dataframe(self, *args, **kwargs):
        """Adds a dataset to a project's staging area (from a pandas DataFrame)."""
        return self.client.add_dataframe(*args, project_id=self.id, **kwargs)

    def commit(self, *args, **kwargs):
        """Adds a commit message to staged resources."""
        return self.client.commit(*args, project_id=self.id, **kwargs)

    def push(self, *args, **kwargs):
        """Pushes the commited resources to the platform."""
        return self.client.push(*args, project_id=self.id, **kwargs)

    def status(self, *args, **kwargs):
        """Shows the state of the staging area."""
        return self.client.status(*args, project_id=self.id, **kwargs)

    def restore(self, *args, **kwargs):
        """Removes the resource specified by ``resource_name`` from the staging area."""
        return self.client.restore(*args, project_id=self.id, **kwargs)



import os
import shutil
import tarfile
import tempfile
import time
import uuid
from typing import List, Optional

import pandas as pd
import yaml

from . import api, exceptions, utils, validators
from .projects import Project
from .schemas import BaselineModelSchema, DatasetSchema, ModelSchema, ShellModelSchema
from .tasks import TaskType
from .version import __version__  # noqa: F401

OPENLAYER_DIR = os.path.join(os.path.expanduser("~"), ".openlayer")
VALID_RESOURCE_NAMES = {"baseline-model", "model", "training", "validation"}


class OpenlayerClient(object):
    """Client class that interacts with the Openlayer Platform.

    Parameters
    ----------
    api_key : str
        Your API key. Retrieve it from the web app.

    Examples
    --------
    Instantiate a client with your api key

    >>> import openlayer
    >>> client = openlayer.OpenlayerClient('YOUR_API_KEY_HERE')
    """

    def __init__(self, api_key: str = None):
        self.api = api.Api(api_key)

        if not os.path.exists(OPENLAYER_DIR):
            os.makedirs(OPENLAYER_DIR)

    def create_project(
        self, name: str, task_type: TaskType, description: Optional[str] = None
    ) -> Project:
        """Creates a project on the Openlayer platform.

        Parameters
        ----------
        name : str
            Name of your project.

            .. important::
                The project name must be unique in a user's collection of projects.

        task_type : :obj:`TaskType`
            Type of ML task. E.g. :obj:`TaskType.TabularClassification` or
            :obj:`TaskType.TextClassification`.

        description : str
            Project description.

        Returns
        -------
        Project
            An object that is used to add models and datasets to the Openlayer platform
            that also contains information about the project.

        Examples
        --------
        Instantiate the client and create the project:

        >>> import openlayer
        >>> client = openlayer.OpenlayerClient('YOUR_API_KEY_HERE')
        >>>
        >>> from openlayer.tasks import TaskType
        >>> project = client.create_project(
        ...     name="Churn prediction",
        ...     task_type=TaskType.TabularClassification,
        ...     description="My first error analysis playground",
        ... )

        With the Project object created, you are able to start adding models and
        datasets to it. Refer to :obj:`add_model` and :obj:`add_dataset` or
        :obj:`add_dataframe` for detailed examples.
        """
        # Validate project
        project_config = {
            "name": name,
            "description": description,
            "task_type": task_type,
        }
        project_validator = validators.ProjectValidator(project_config=project_config)
        failed_validations = project_validator.validate()

        if failed_validations:
            raise exceptions.OpenlayerValidationError(
                "There are issues with the project. \n"
                "Make sure to fix all of the issues listed above before creating it.",
            ) from None

        endpoint = "projects"
        payload = dict(name=name, description=description, taskType=task_type.value)
        project_data = self.api.post_request(endpoint, body=payload)

        project = Project(project_data, self.api.upload, self)

        # Check if the staging area exists
        project_dir = os.path.join(OPENLAYER_DIR, f"{project.id}/staging")
        os.makedirs(project_dir)

        print(f"Created your project. Navigate to {project.links['app']} to see it.")
        return project

    def load_project(self, name: str) -> Project:
        """Loads an existing project from the Openlayer platform.

        Parameters
        ----------
        name : str
            Name of the project to be loaded. The name of the project is the one
            displayed on the Openlayer platform.

            .. note::
                If you haven't created the project yet, you should use the
                :obj:`create_project` method.

        Returns
        -------
        Project
            An object that is used to add models and datasets to the Openlayer platform
            that also contains information about the project.

        Examples
        --------
        Instantiate the client and load the project:

        >>> import openlayer
        >>> client = openlayer.OpenlayerClient('YOUR_API_KEY_HERE')
        >>>
        >>> project = client.load_project(name="Churn prediction")

        With the Project object loaded, you are able to add models and datasets to
        the it. Refer to :obj:`add_model` and :obj:`add_dataset` or
        :obj:`add_dataframe` for detailed examples.
        """
        endpoint = f"projects?name={name}"
        project_data = self.api.get_request(endpoint)
        if len(project_data["items"]) == 0:
            raise exceptions.OpenlayerResourceNotFound(
                f"Project with name {name} not found."
            )
        project = Project(project_data["items"][0], self.api.upload, self)

        # Create the project staging area, if it doesn't yet exist
        project_dir = os.path.join(OPENLAYER_DIR, f"{project.id}/staging")
        if not os.path.exists(project_dir):
            os.makedirs(project_dir)

        print(f"Found your project. Navigate to {project.links['app']} to see it.")
        return project

    def create_or_load_project(
        self, name: str, task_type: TaskType, description: Optional[str] = None
    ) -> Project:
        """Helper function that returns a project given a name. Creates a new project
        if no project with the name exists.

        Parameters
        ----------
        name : str
            Name of your project.

            .. important::
                The project name must be unique in a user's collection of projects.

        task_type : :obj:`TaskType`
            Type of ML task. E.g. :obj:`TaskType.TabularClassification` or
            :obj:`TaskType.TextClassification`.

        description : str
            Project description.

        Returns
        -------
        Project
            An object that is used to add models and datasets to the Openlayer platform
            that also contains information about the project.

        Examples
        --------
        Instantiate the client and create or load the project:

        >>> import openlayer
        >>> client = openlayer.OpenlayerClient('YOUR_API_KEY_HERE')
        >>>
        >>> from openlayer.tasks import TaskType
        >>> project = client.create_or_load_project(
        ...     name="Churn prediction",
        ...     task_type=TaskType.TabularClassification,
        ...     description="My first error analysis playground",
        ... )

        With the Project object, you are able to start adding models and
        datasets to it. Refer to :obj:`add_model` and :obj:`add_dataset` or
        :obj:`add_dataframe` for detailed examples.
        """
        try:
            return self.load_project(name)
        except exceptions.OpenlayerResourceNotFound:
            return self.create_project(
                name=name, task_type=task_type, description=description
            )

    def add_model(
        self,
        model_config_file_path: str,
        model_package_dir: Optional[str] = None,
        sample_data: Optional[pd.DataFrame] = None,
        force: bool = False,
        project_id: str = None,
    ):
        """Adds a model to a project's staging area.

        Parameters
        ----------
        model_config_file_path : str
            Path to the model configuration YAML file.

            .. admonition:: What's on the model config file?

                The content of the YAML file depends on whether you are adding a shell
                model or a model package.

                **If you are adding a shell model**, the model configuration file
                must contain the following fields:

                - ``name`` : str
                    Name of the model.
                - ``architectureType`` : str
                    The model's framework. Must be one of the supported frameworks
                    on :obj:`ModelType`.
                - ``metadata`` : Dict[str, any], default {}
                    Dictionary containing metadata about the model. This is the
                    metadata that will be displayed on the Openlayer platform.

                **Alternatively, if you are adding a model package** (i.e., with model
                artifacts and prediction interface), the model configuration file must
                contain the following fields:

                - ``name`` : str
                    Name of the model.
                - ``architectureType`` : str
                    The model's framework. Must be one of the supported frameworks
                    on :obj:`ModelType`.
                - ``classNames`` : List[str]
                    List of class names corresponding to the outputs of your predict function.
                    E.g. ``['positive', 'negative']``.
                - ``featureNames`` : List[str], default []
                    List of input feature names. Only applicable if your ``task_type`` is
                    :obj:`TaskType.TabularClassification` or :obj:`TaskType.TabularRegression`.
                - ``categoricalFeatureNames`` : List[str], default []
                    A list containing the names of all categorical features used by the model.
                    E.g. ``["Gender", "Geography"]``. Only applicable if your ``task_type`` is
                    :obj:`TaskType.TabularClassification` or :obj:`TaskType.TabularRegression`.
                - ``metadata`` : Dict[str, any], default {}
                    Dictionary containing metadata about the model. This is the metadata that
                    will be displayed on the Openlayer platform.


        model_package_dir : str, default None
            Path to the directory containing the model package. **Only needed if you are
            interested in adding the model's artifacts.**

            .. admonition:: What's inside `model_package_dir`?

                The model package directory must contain the following files:

                - ``prediction_interface.py``
                    The prediction interface file.
                - ``model artifacts``
                    The model artifacts. This can be a single file or a directory containing
                    multiple files. The model artifacts must be compatible with the
                    prediction interface file.
                - ``requirements.txt``
                    The requirements file. This file contains the dependencies needed to run
                    the prediction interface file.

                For instructions on how to create a model package, refer to
                the documentation.

        sample_data : pd.DataFrame, default None
            Sample data that can be run through the model. **Only needed if  model_package_dir
            is not None**. This data is used to ensure
            the model's prediction interface is compatible with the Openlayer platform.

            .. important::
                The sample_data must be a dataframe with at least two rows.
        force : bool
            If :obj:`add_model` is called when there is already a model in the staging area,
            when ``force=True``, the existing staged model will be overwritten by the new
            one. When ``force=False``, the user will be prompted to confirm the
            overwrite.

        Examples
        --------

        .. seealso::
            Our `sample notebooks
            <https://github.com/openlayer-ai/openlayer-python/tree/main/examples>`_ and
            `tutorials <https://docs.openlayer.com/docs/overview-of-tutorial-tracks>`_.

        First, instantiate the client:

        >>> import openlayer
        >>> client = openlayer.OpenlayerClient('YOUR_API_KEY_HERE')

        Create a project if you don't have one:

        >>> from openlayer.tasks import TaskType
        >>> project = client.create_project(
        ...     name="Churn Prediction",
        ...     task_type=TaskType.TabularClassification,
        ...     description="My first project!",
        ... )

        If you already have a project created on the platform:

        >>> project = client.load_project(name="Your project name")

        **If your project's task type is tabular classification...**

        Let's say your model expects to receive a dataset looks like
        the following as input:

        >>> df
            CreditScore  Geography    Balance
        0           618     France     321.92
        1           714    Germany  102001.22
        2           604      Spain   12333.15
        ..          ...        ...        ...

        Then, you can add the model to the project with:

        >>> project.add_model(
        ...     model_config_file_path="path/to/model/config",
        ...     model_package_dir="path/to/model/package")
        ...     sample_data=df.iloc[:5, :],
        ... )

        After adding the model to the project, it is staged, waiting to
        be committed and pushed to the platform. You can check what's on
        your staging area with :obj:`status`. If you want to push the model
        right away with a commit message, you can use the :obj:`commit` and
        :obj:`push` methods:

        >>> project.commit("Initial model commit.")
        >>> project.push()

        **If your task type is text classification...**

        Let's say your dataset looks like the following:

        >>> df
                                      Text
        0    I have had a long weekend
        1    I'm in a fantastic mood today
        2    Things are looking up
        ..                             ...

        Then, you can add the model to the project with:

        >>> project.add_model(
        ...     model_config_file_path="path/to/model/config",
        ...     model_package_dir="path/to/model/package")
        ...     sample_data=df.iloc[:5, :],
        ... )

        After adding the model to the project, it is staged, waiting to
        be committed and pushed to the platform. You can check what's on
        your staging area with :obj:`status`. If you want to push the model
        right away with a commit message, you can use the :obj:`commit` and
        :obj:`push` methods:

        >>> project.commit("Initial model commit.")
        >>> project.push()
        """
        # Basic argument combination checks
        if (model_package_dir is not None and sample_data is None) or (
            model_package_dir is None and sample_data is not None
        ):
            raise ValueError(
                "Both `model_package_dir` and `sample_data` must be provided together to"
                " add a model with its artifacts to the platform."
            )
        if sample_data is not None:
            if not isinstance(sample_data, pd.DataFrame):
                raise ValueError(
                    "The sample data must be a pandas DataFrame with at least 2 rows."
                )
            elif len(sample_data) < 2:
                raise ValueError(
                    "The sample data must contain at least 2 rows, but only"
                    f"{len(sample_data)} rows were provided."
                )

        # Validate model package
        model_package_validator = validators.ModelValidator(
            model_package_dir=model_package_dir,
            model_config_file_path=model_config_file_path,
            sample_data=sample_data,
        )
        failed_validations = model_package_validator.validate()

        if failed_validations:
            raise exceptions.OpenlayerValidationError(
                "There are issues with the model package. \n"
                "Make sure to fix all of the issues listed above before the upload.",
            ) from None

        # Load model config and augment with defaults
        model_config = utils.read_yaml(model_config_file_path)
        if model_package_dir:
            model_data = ModelSchema().load(model_config)
        else:
            model_data = ShellModelSchema().load(model_config)

        # Copy relevant resources to temp directory
        with tempfile.TemporaryDirectory() as temp_dir:
            if model_package_dir:
                shutil.copytree(model_package_dir, temp_dir, dirs_exist_ok=True)
                utils.write_python_version(temp_dir)

            utils.write_yaml(model_data, f"{temp_dir}/model_config.yaml")

            self._stage_resource(
                resource_name="model",
                resource_dir=temp_dir,
                project_id=project_id,
                force=force,
            )

    def add_baseline_model(
        self,
        project_id: int,
        task_type: TaskType,
        model_config_file_path: Optional[str] = None,
        force: bool = False,
    ):
        """
        **Coming soon...**

        Adds a baseline model to the project.

        Baseline models should be added together with training and validation
        sets. A model will then be trained on the platform using AutoML, using
        the parameters provided in the model config file.

        .. important::
            This feature is experimental and currently under development. Only
            tabular classification tasks are supported for now.

        Parameters
        ----------
        model_config_file_path : str, optional
            Path to the model configuration YAML file. If not provided, the default
            model config will be used.

            .. admonition:: What's on the model config file?

                For baseline models, the content of the YAML file should contain:

                - ``ensembleSize`` : int, default 10
                    Number of models ensembled.
                - ``randomSeed`` : int, default 42
                    Random seed to be used for model training.
                - ``timeout`` : int, default 60
                    Maximum time (in seconds) to train all the models.
                - ``perRunLimit`` : int, optional
                    Maximum time (in seconds) to train each model.
                - ``metadata`` : Dict[str, any], default {}
                    Dictionary containing metadata about the model. This is the
                    metadata that will be displayed on the Openlayer platform.
        force : bool, optional
            Whether to force the addition of the baseline model to the project.
            If set to True, any existing staged baseline model will be overwritten.
        """
        if task_type is not TaskType.TabularClassification:
            raise exceptions.OpenlayerException(
                "Only tabular classification is supported for model baseline for now."
            )

        # Validate the baseline model
        baseline_model_validator = validators.BaselineModelValidator(
            model_config_file_path=model_config_file_path,
        )
        failed_validations = baseline_model_validator.validate()

        if failed_validations:
            raise exceptions.OpenlayerValidationError(
                "There are issues with the baseline model. \n"
                "Make sure to fix all of the issues listed above before the upload.",
            ) from None

        # Load model config and augment with defaults
        model_config = {}
        if model_config_file_path is not None:
            model_config = utils.read_yaml(model_config_file_path)
        model_data = BaselineModelSchema().load(model_config)

        # Copy relevant resources to temp directory
        with tempfile.TemporaryDirectory() as temp_dir:
            utils.write_yaml(model_data, f"{temp_dir}/model_config.yaml")

            self._stage_resource(
                resource_name="baseline-model",
                resource_dir=temp_dir,
                project_id=project_id,
                force=force,
            )

    def add_dataset(
        self,
        file_path: str,
        dataset_config_file_path: str,
        project_id: str = None,
        force: bool = False,
    ):
        r"""Adds a dataset to a project's staging area (from a csv).

        Parameters
        ----------
        file_path : str
            Path to the csv file containing the dataset.
        dataset_config_file_path : str
            Path to the dataset configuration YAML file.

            .. admonition:: What's on the dataset config file?

                The YAML file with the dataset config must have the following fields:

                - ``columnNames`` : List[str]
                    List of the dataset's column names.
                - ``classNames`` : List[str]
                    List of class names indexed by label integer in the dataset.
                    E.g. ``[negative, positive]`` when ``[0, 1]`` are in your label column.
                - ``labelColumnName`` : str
                    Column header in the csv containing the labels.

                    .. important::
                        The labels in this column must be zero-indexed integer values.
                - ``label`` : str
                    Type of dataset. E.g. ``'training'`` or
                    ``'validation'``.
                - ``featureNames`` : List[str], default []
                    List of input feature names. Only applicable if your ``task_type`` is
                    :obj:`TaskType.TabularClassification` or :obj:`TaskType.TabularRegression`.
                - ``textColumnName`` : str, default None
                    Column header in the csv containing the input text. Only applicable if
                    your ``task_type`` is :obj:`TaskType.TextClassification`.
                - ``predictionsColumnName`` : str, default None
                    Column header in the csv containing the predictions. Only applicable if you
                    are uploading a model as well with the :obj:`add_model` method.

                    .. important::
                        Each cell in this column must contain a list of
                        class probabilities. For example, for a binary classification
                        task, the cell values should look like this:
                        .. csv-table::
                            :header: ..., predictions
                            ..., "[0.6650292861587155, 0.3349707138412845]"
                            ..., "[0.8145561636482788, 0.18544383635172124]"

                - ``categoricalFeatureNames`` : List[str], default []
                    A list containing the names of all categorical features in the dataset.
                    E.g. ``["Gender", "Geography"]``. Only applicable if your ``task_type`` is
                    :obj:`TaskType.TabularClassification` or :obj:`TaskType.TabularRegression`.
                - ``language`` : str, default 'en'
                    The language of the dataset in ISO 639-1 (alpha-2 code) format.
                - ``sep`` : str, default ','
                    Delimiter to use. E.g. `'\\t'`.
        force : bool
            If :obj:`add_dataset` is called when there is already a dataset of the same type
            in the staging area, when ``force=True``, the existing staged dataset will be
            overwritten by the new one. When ``force=False``, the user will be prompted
            to confirm the overwrite.

        Notes
        -----
        - Please ensure your input features are strings, ints or floats.
        - Please ensure your label column name is not contained in ``feature_names``.

        Examples
        --------

        First, instantiate the client:

        >>> import openlayer
        >>> client = openlayer.OpenlayerClient('YOUR_API_KEY_HERE')

        Create a project if you don't have one:

        >>> from openlayer.tasks import TaskType
        >>> project = client.create_project(
        ...     name="Churn Prediction",
        ...     task_type=TaskType.TabularClassification,
        ...     description="My first project!",
        ... )

        If you already have a project created on the platform:

        >>> project = client.load_project(name="Your project name")

        **If your project's task type is tabular classification...**

        Let's say your dataset looks like the following:

        .. csv-table::
            :header: CreditScore, Geography, Balance, Churned

            618, France, 321.92, 1
            714, Germany, 102001.22, 0
            604, Spain, 12333.15, 0

        .. important::
            The labels in your csv **must** be integers that correctly index into the
            ``class_names`` array that you define (as shown below).
            E.g. 0 => 'Retained', 1 => 'Churned'

        Write the dataset config YAML file with the variables are needed by Openlayer:

        >>> import yaml
        >>>
        >> dataset_config = {
        ...     'columnNames': ['CreditScore', 'Geography', 'Balance', 'Churned'],
        ...     'classNames': ['Retained', 'Churned'],
        ...     'labelColumnName': 'Churned',
        ...     'label': 'training',  # or 'validation'
        ...     'featureNames': ['CreditScore', 'Geography', 'Balance'],
        ...     'categoricalFeatureNames': ['Geography'],
        ... }
        >>>
        >>> with open('/path/to/dataset_config.yaml', 'w') as f:
        ...     yaml.dump(dataset_config, f)

        You can now add this dataset to your project with:

        >>> project.add_dataset(
        ...     file_path='/path/to/dataset.csv',
        ...     dataset_config_file_path='/path/to/dataset_config.yaml',
        ... )

        After adding the dataset to the project, it is staged, waiting to
        be committed and pushed to the platform. You can check what's on
        your staging area with :obj:`status`. If you want to push the dataset
        right away with a commit message, you can use the :obj:`commit` and
        :obj:`push` methods:

        >>> project.commit("Initial dataset commit.")
        >>> project.push()

        **If your task type is text classification...**

        Let's say your dataset looks like the following:

        .. csv-table::
            :header: Text, Sentiment

            I have had a long weekend, 0
            I'm in a fantastic mood today, 1
            Things are looking up, 1

        Write the dataset config YAML file with the variables are needed by Openlayer:

        >>> import yaml
        >>>
        >> dataset_config = {
        ...     'columnNames': ['Text', 'Sentiment'],
        ...     'classNames': ['Negative', 'Positive'],
        ...     'labelColumnName': 'Sentiment',
        ...     'label': 'training',  # or 'validation'
        ...     'textColumnName': 'Text',
        ... }
        >>>
        >>> with open('/path/to/dataset_config.yaml', 'w') as f:
        ...     yaml.dump(dataset_config, f)

        You can now add this dataset to your project with:

        >>> project.add_dataset(
        ...     file_path='/path/to/dataset.csv',
        ...     dataset_config_file_path='/path/to/dataset_config.yaml',
        ... )

        After adding the dataset to the project, it is staged, waiting to
        be committed and pushed to the platform. You can check what's on
        your staging area with :obj:`status`. If you want to push the dataset
        right away with a commit message, you can use the :obj:`commit` and
        :obj:`push` methods:

        >>> project.commit("Initial dataset commit.")
        >>> project.push()
        """
        # Validate dataset
        dataset_validator = validators.DatasetValidator(
            dataset_config_file_path=dataset_config_file_path,
            dataset_file_path=file_path,
        )
        failed_validations = dataset_validator.validate()

        if failed_validations:
            raise exceptions.OpenlayerValidationError(
                "There are issues with the dataset and its config. \n"
                "Make sure to fix all of the issues listed above before the upload.",
            ) from None

        # Load dataset config and augment with defaults
        dataset_config = utils.read_yaml(dataset_config_file_path)
        dataset_data = DatasetSchema().load(dataset_config)

        # Copy relevant resources to temp directory
        with tempfile.TemporaryDirectory() as temp_dir:
            shutil.copy(file_path, f"{temp_dir}/dataset.csv")
            utils.write_yaml(dataset_data, f"{temp_dir}/dataset_config.yaml")

            self._stage_resource(
                resource_name=dataset_data.get("label"),
                resource_dir=temp_dir,
                project_id=project_id,
                force=force,
            )

    def add_dataframe(
        self,
        dataset_df: pd.DataFrame,
        dataset_config_file_path: str,
        project_id: str = None,
        force: bool = False,
    ):
        r"""Adds a dataset to a project's staging area (from a pandas DataFrame).

        Parameters
        ----------
        dataset_df : pd.DataFrame
            Dataframe containing your dataset.
        dataset_config_file_path : str
            Path to the dataset configuration YAML file.

            .. admonition:: What's on the dataset config file?

                The YAML file with the dataset config must have the following fields:

                - ``columnNames`` : List[str]
                    List of the dataset's column names.
                - ``classNames`` : List[str]
                    List of class names indexed by label integer in the dataset.
                    E.g. ``[negative, positive]`` when ``[0, 1]`` are in your label column.
                - ``labelColumnName`` : str
                    Column header in the csv containing the labels.

                    .. important::
                        The labels in this column must be zero-indexed integer values.
                - ``label`` : str
                    Type of dataset. E.g. ``'training'`` or
                    ``'validation'``.
                - ``featureNames`` : List[str], default []
                    List of input feature names. Only applicable if your ``task_type`` is
                    :obj:`TaskType.TabularClassification` or :obj:`TaskType.TabularRegression`.
                - ``textColumnName`` : str, default None
                    Column header in the csv containing the input text. Only applicable if your
                    ``task_type`` is :obj:`TaskType.TextClassification`.
                - ``predictionsColumnName`` : str, default None
                    Column header in the csv containing the predictions. Only applicable if you
                    are uploading a model as well with the :obj:`add_model` method.

                    .. important::
                        Each cell in this column must contain a list of
                        class probabilities. For example, for a binary classification
                        task, the cell values should look like this:
                        .. csv-table::
                            :header: ..., predictions
                            ..., "[0.6650292861587155, 0.3349707138412845]"
                            ..., "[0.8145561636482788, 0.18544383635172124]"

                - ``categoricalFeatureNames`` : List[str], default []
                    A list containing the names of all categorical features in the dataset.
                    E.g. ``["Gender", "Geography"]``. Only applicable if your ``task_type`` is
                    :obj:`TaskType.TabularClassification` or :obj:`TaskType.TabularRegression`.
                - ``language`` : str, default 'en'
                    The language of the dataset in ISO 639-1 (alpha-2 code) format.
                - ``sep`` : str, default ','
                    Delimiter to use. E.g. `'\\t'`.
        force : bool
            If :obj:`add_dataframe` is called when there is already a dataset of the same
            type in the staging area, when ``force=True``, the existing staged dataset will
            be overwritten by the new one. When ``force=False``, the user will be prompted
            to confirm the overwrite.

        Notes
        -----
        - Please ensure your input features are strings, ints or floats.
        - Please ensure your label column name is not contained in ``feature_names``.

        Examples
        --------

        First, instantiate the client:

        >>> import openlayer
        >>> client = openlayer.OpenlayerClient('YOUR_API_KEY_HERE')

        Create a project if you don't have one:

        >>> from openlayer.tasks import TaskType
        >>> project = client.create_project(
        ...     name="Churn Prediction",
        ...     task_type=TaskType.TabularClassification,
        ...     description="My first project!",
        ... )

        If you already have a project created on the platform:

        >>> project = client.load_project(name="Your project name")

        **If your project's task type is tabular classification...**

        Let's say your dataframe looks like the following:

        >>> df
            CreditScore  Geography    Balance  Churned
        0           618     France     321.92        1
        1           714    Germany  102001.22        0
        2           604      Spain   12333.15        0

        .. important::
            The labels in your dataframe **must** be integers that correctly index into
            the ``class_names`` array that you define (as shown below).
            E.g. 0 => 'Retained', 1 => 'Churned'.

        Write the dataset config YAML file with the variables are needed by Openlayer:

        >>> import yaml
        >>>
        >> dataset_config = {
        ...     'columnNames': ['CreditScore', 'Geography', 'Balance', 'Churned'],
        ...     'classNames': ['Retained', 'Churned'],
        ...     'labelColumnName': 'Churned',
        ...     'label': 'training',  # or 'validation'
        ...     'featureNames': ['CreditScore', 'Geography', 'Balance'],
        ...     'categoricalFeatureNames': ['Geography'],
        ... }
        >>>
        >>> with open('/path/to/dataset_config.yaml', 'w') as f:
        ...     yaml.dump(dataset_config, f)

        You can now add this dataset to your project with:

        >>> project.add_dataframe(
        ...     dataset_df=df,
        ...     dataset_config_file_path='/path/to/dataset_config.yaml',
        ... )

        After adding the dataset to the project, it is staged, waiting to
        be committed and pushed to the platform. You can check what's on
        your staging area with :obj:`status`. If you want to push the dataset
        right away with a commit message, you can use the :obj:`commit` and
        :obj:`push` methods:

        >>> project.commit("Initial dataset commit.")
        >>> project.push()

        **If your task type is text classification...**

        Let's say your dataset looks like the following:

        >>> df
                                      Text  Sentiment
        0    I have had a long weekend              0
        1    I'm in a fantastic mood today          1
        2    Things are looking up                  1

        Write the dataset config YAML file with the variables are needed by Openlayer:

        >>> import yaml
        >>>
        >> dataset_config = {
        ...     'columnNames': ['Text', 'Sentiment'],
        ...     'classNames': ['Negative', 'Positive'],
        ...     'labelColumnName': 'Sentiment',
        ...     'label': 'training',  # or 'validation'
        ...     'textColumnName': 'Text',
        ... }
        >>>
        >>> with open('/path/to/dataset_config.yaml', 'w') as f:
        ...     yaml.dump(dataset_config, f)

        You can now add this dataset to your project with:

        >>> project.add_dataframe(
        ...     dataset_df=df,
        ...     dataset_config_file_path='/path/to/dataset_config.yaml',
        ... )

        After adding the dataset to the project, it is staged, waiting to
        be committed and pushed to the platform. You can check what's on
        your staging area with :obj:`status`. If you want to push the dataset
        right away with a commit message, you can use the :obj:`commit` and
        :obj:`push` methods:

        >>> project.commit("Initial dataset commit.")
        >>> project.push()
        """
        # --------------------------- Resource validations --------------------------- #
        if not isinstance(dataset_df, pd.DataFrame):
            raise exceptions.OpenlayerValidationError(
                f"- `dataset_df` is a `{type(dataset_df)}`, but it must be of type"
                " `pd.DataFrame`. \n"
            ) from None
        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = os.path.join(tmp_dir, str(uuid.uuid1()))
            dataset_df.to_csv(file_path, index=False)
            return self.add_dataset(
                file_path=file_path,
                project_id=project_id,
                dataset_config_file_path=dataset_config_file_path,
                force=force,
            )

    def commit(self, message: str, project_id: int, force: bool = False):
        """Adds a commit message to staged resources.

        Parameters
        ----------
        message : str
            The commit message, between 1 and 140 characters.
        force : bool
            If :obj:`commit` is called when there is already a commit message,
            when ``force=True``, the existing message will be overwritten by the new
            one. When ``force=False``, the user will be prompted to confirm the overwrite.

        Notes
        -----
        - To use this method, you must first add a model and/or dataset to the staging area using one of the ``add`` methods (e.g., :obj:`add_model`, :obj:`add_dataset`, :obj:`add_dataframe`).

        Examples
        --------
        A commit message is associated with a project version. We have a new project version
        every time any of its resources (namely, model and/or dataset) are updated. The commit
        message is supposed to be a short description of the changes from one version to the next.

        Let's say you have a project with a model and a dataset staged. You can confirm these
        resources are indeed in the staging area using the :obj:`status` method:

        >>> project.status()

        Now, you can add a commit message to the staged resources.

        >>> project.commit("Initial commit.")

        After adding the commit message, the resources are ready to be pushed to the platform.
        You use the :obj:`push` method to do so:

        >>> project.push()
        """
        # Validate commit
        commit_validator = validators.CommitValidator(commit_message=message)
        failed_validations = commit_validator.validate()

        if failed_validations:
            raise exceptions.OpenlayerValidationError(
                "There are issues with the commit message specified. \n"
                "Make sure to fix all of the issues listed above before committing.",
            ) from None

        project_dir = f"{OPENLAYER_DIR}/{project_id}/staging"

        if not os.listdir(project_dir):
            print(
                "There is nothing staged to commit. Please add model and/or datasets"
                " first before committing."
            )
            return

        if os.path.exists(f"{project_dir}/commit.yaml"):
            print("Found a previous commit that was not pushed to the platform.")
            overwrite = "n"

            if not force:
                with open(
                    f"{project_dir}/commit.yaml", "r", encoding="UTF-8"
                ) as commit_file:
                    commit = yaml.safe_load(commit_file)
                    print(
                        f"\t - Commit message: `{commit['message']}` \n \t - Date: {commit['date']}"
                    )
                overwrite = input(
                    "Do you want to overwrite it with the current message? [y/n]: "
                )
            if overwrite.lower() == "y" or force:
                print("Overwriting commit message...")
                os.remove(f"{project_dir}/commit.yaml")

            else:
                print("Keeping the existing commit message.")
                return

        commit = dict(message=message, date=time.ctime())
        with open(f"{project_dir}/commit.yaml", "w", encoding="UTF-8") as commit_file:
            yaml.dump(commit, commit_file)

        print("Committed!")

    def push(self, project_id: int):
        """Pushes the commited resources to the platform.

        Notes
        -----
        - To use this method, you must first have committed your changes with the :obj:`commit`
        method.

        Examples
        --------

        Let's say you have a project with a model and a dataset staged and committed. You can
        confirm these resources are indeed in the staging area using the :obj:`status` method:

        >>> project.status()

        You should see the staged resources as well as the commit message associated with them.

        Now, you can push the resources to the platform with:

        >>> project.push()
        """
        project_dir = f"{OPENLAYER_DIR}/{project_id}/staging"

        if not os.listdir(project_dir):
            print(
                "The staging area is clean and there is nothing committed to push. "
                "Please add model and/or datasets first, and then commit before pushing."
            )
            return

        if not os.path.exists(f"{project_dir}/commit.yaml"):
            print(
                "There are resources staged, but you haven't committed them yet. "
                "Please commit before pushing"
            )
            return

        with open(f"{project_dir}/commit.yaml", "r", encoding="UTF-8") as commit_file:
            commit = yaml.safe_load(commit_file)

        # Validate bundle resources
        commit_bundle_validator = validators.CommitBundleValidator(
            bundle_path=project_dir,
            skip_dataset_validation=True,
            skip_model_validation=False,  # Don't skip because the sample data is different
        )
        failed_validations = commit_bundle_validator.validate()

        if failed_validations:
            raise exceptions.OpenlayerValidationError(
                "There are issues with the staged resources. \n"
                "Make sure to fix all of the issues listed above before pushing.",
            ) from None

        print(
            "Pushing changes to the platform with the commit message: \n"
            f"\t - Message: {commit['message']} \n"
            f"\t - Date: {commit['date']}"
        )

        # Tar the project's staging area
        with tempfile.TemporaryDirectory() as tmp_dir:
            tar_file_path = os.path.join(tmp_dir, "staging")
            with tarfile.open(tar_file_path, mode="w:gz") as tar:
                tar.add(project_dir, arcname=os.path.basename(project_dir))

            # Upload the tar file
            payload = {"commit": {"message": commit["message"]}}
            self.api.upload(
                endpoint=f"projects/{project_id}/versions",
                file_path=tar_file_path,
                object_name="tarfile",
                body=payload,
            )

        # Clean up the staging area
        shutil.rmtree(project_dir)
        os.makedirs(project_dir, exist_ok=True)

        print("Pushed!")

    def status(self, project_id: int):
        """Shows the state of the staging area.

        Examples
        --------

        You can use the :obj:`status` method to check the state of the staging area.

        >>> project.status()

        The staging area can be in one of three states.

        You can have a clean staging area, which is the initial state as well as the state
        after you have pushed your changes to the platform (with the :obj:`push` method).

        You can have a staging area with different resources staged (e.g., models and datasets
        added with the :obj:`add_model`, :obj:`add_dataset`, and :obj:`add_dataframe` mehtods).

        Finally, you can have a staging area with resources staged and committed (with the
        :obj:`commit` method).
        """
        project_dir = f"{OPENLAYER_DIR}/{project_id}/staging"

        if not os.listdir(project_dir):
            print(
                "The staging area is clean. You can stage models and/or datasets by"
                " using the corresponding `add` methods."
            )
            return

        if not os.path.exists(f"{project_dir}/commit.yaml"):
            print("The following resources are staged, waiting to be committed:")
            for file in os.listdir(project_dir):
                if file in VALID_RESOURCE_NAMES:
                    print(f"\t - {file}")
            print("Use the `commit` method to add a commit message to your changes.")
            return

        with open(f"{project_dir}/commit.yaml", "r", encoding="UTF-8") as commit_file:
            commit = yaml.safe_load(commit_file)
        print("The following resources are committed, waiting to be pushed:")
        for file in os.listdir(project_dir):
            if file in VALID_RESOURCE_NAMES:
                print(f"\t - {file}")
        print(f"Commit message from {commit['date']}:")
        print(f"\t {commit['message']}")
        print("Use the `push` method to push your changes to the platform.")

    def restore(self, *resource_names: str, project_id: int):
        """Removes the resource specified by ``resource_name`` from the staging area.

        Parameters
        ----------
        *resource_names : str
            The names of the resources to restore, separated by comma. Valid resource names
            are ``"model"``, ``"training"``, and ``"validation"``.

            .. important::
                To see the names of the resources staged, use the :obj:`status` method.

        Examples
        --------
        Let's say you have initially used the :obj:`add_model` method to add a model to the
        staging area.

        >>> project.add_model(
        ...     model_package_dir="/path/to/model/package",
        ...     sample_data=df
        ... )

        You can see the model staged with the :obj:`status` method:

        >>> project.status()

        You can then remove the model from the staging area with the :obj:`restore` method:

        >>> project.restore(resource_name="model")
        """
        project_dir = f"{OPENLAYER_DIR}/{project_id}/staging"

        for resource_name in resource_names:
            if not os.path.exists(f"{project_dir}/{resource_name}"):
                print(
                    f"There's no resource named `{resource_name}` in the staging area. "
                    "Make sure that you are trying to restore a staged resource. "
                    "To see the names of the resources staged, use the `status` method."
                )
                return

            shutil.rmtree(f"{project_dir}/{resource_name}")
            print(f"Removed resource `{resource_name}` from the staging area.")

            # Remove commit if there are no more resources staged
            if len(os.listdir(project_dir)) == 1 and os.path.exists(
                f"{project_dir}/commit.yaml"
            ):
                os.remove(f"{project_dir}/commit.yaml")

    def _stage_resource(
        self, resource_name: str, resource_dir: str, project_id: int, force: bool
    ):
        """Adds the resource specified by `resource_name` to the project's staging directory.

        Parameters
        ----------
        resource_name : str
            The name of the resource to stage. Can be one of "model", "training",
            or "validation".
        resource_dir : str
            The path from which to copy the resource.
        project_id : int
            The id of the project to which the resource should be added.
        force : bool
            Whether to overwrite the resource if it already exists in the staging area.
        """
        if resource_name not in VALID_RESOURCE_NAMES:
            raise ValueError(
                "Resource name must be one of 'baseline-model', 'model', 'training', or"
                f" 'validation', but got '{resource_name}'."
            )

        project_dir = f"{OPENLAYER_DIR}/{project_id}/staging"

        resources_staged = utils.list_resources_in_bundle(project_dir)

        if resource_name == "model" and "baseline-model" in resources_staged:
            raise exceptions.OpenlayerException(
                "Trying to stage a `model` when there is a `baseline-model` already staged."
                + " You can either add a `model` or a `baseline-model`, but not both at the"
                + " same time. Please remove one of them from the staging area using the"
                + " `restore` method."
            ) from None

        if resource_name == "baseline-model" and "model" in resources_staged:
            raise exceptions.OpenlayerException(
                "Trying to stage a `baseline-model` when there is a `model` already staged."
                + " You can either add a `model` or a `baseline-model`, but not both at the"
                + " same time. Please remove one of them from the staging area using the"
                + " `restore` method."
            ) from None

        if resource_name in resources_staged:
            print(f"Found an existing `{resource_name}` resource staged.")

            overwrite = "n"
            if not force:
                overwrite = input("Do you want to overwrite it? [y/n] ")
            if overwrite.lower() == "y" or force:
                print(f"Overwriting previously staged `{resource_name}` resource...")
                shutil.rmtree(project_dir + "/" + resource_name)
            else:
                print(f"Keeping the existing `{resource_name}` resource staged.")
                return

        shutil.copytree(resource_dir, project_dir + "/" + resource_name)

        print(f"Staged the `{resource_name}` resource!")
